import uuid
from contextlib import contextmanager
from logging import Logger
from typing import Iterator

from layerapi.api.ids_pb2 import AccountId
from layerapi.api.service.account.account_api_pb2 import (
    GetAccountViewByIdRequest,
    GetAccountViewByIdResponse,
)
from layerapi.api.service.account.account_api_pb2_grpc import AccountAPIStub
from layerapi.api.service.account.user_api_pb2 import GetMyOrganizationRequest
from layerapi.api.service.account.user_api_pb2_grpc import UserAPIStub

from layer.config import ClientConfig
from layer.contracts.accounts import Account
from layer.utils.grpc import create_grpc_channel


class AccountServiceClient:
    _service: UserAPIStub
    _account_api: AccountAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config.account_service
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["AccountServiceClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = UserAPIStub(channel=channel)
            self._account_api = AccountAPIStub(channel=channel)
            yield self

    def get_account_name_by_id(self, account_id: uuid.UUID) -> str:
        get_my_org_resp: GetAccountViewByIdResponse = (
            self._account_api.GetAccountViewById(
                GetAccountViewByIdRequest(id=AccountId(value=str(account_id))),
            )
        )
        return get_my_org_resp.account_view.name

    def get_my_account(self) -> Account:
        # TODO LAY-2692
        org = self._service.GetMyOrganization(
            GetMyOrganizationRequest(),
        ).organization
        return Account(
            id=uuid.UUID(org.id.value),
            name=org.name,
        )



from contextlib import contextmanager
from logging import Logger
from typing import Iterator, List, Optional, cast
from uuid import UUID

from layerapi.api.entity.logged_model_metric_pb2 import LoggedModelMetric
from layerapi.api.ids_pb2 import DatasetBuildId, ModelMetricId, ModelTrainId
from layerapi.api.service.logged_data.logged_data_api_pb2 import (
    GetLoggedDataRequest,
    LogDataRequest,
    LogDataResponse,
    LogModelMetricRequest,
)
from layerapi.api.service.logged_data.logged_data_api_pb2_grpc import LoggedDataAPIStub
from layerapi.api.value.logged_data_type_pb2 import LoggedDataType

from layer.config import ClientConfig
from layer.contracts.logged_data import LoggedData
from layer.contracts.logged_data import LoggedDataType as LDType
from layer.contracts.logged_data import ModelMetricPoint
from layer.utils.grpc import create_grpc_channel


class LoggedDataClient:
    _service: LoggedDataAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
    ):
        self._config = config
        self._logger = logger
        self._access_token = config.access_token
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path

    @contextmanager
    def init(self) -> Iterator["LoggedDataClient"]:
        with create_grpc_channel(
            self._config.grpc_gateway_address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = LoggedDataAPIStub(channel=channel)
            yield self

    def get_logged_data(
        self,
        tag: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> LoggedData:
        request = GetLoggedDataRequest(
            unique_tag=tag,
            model_train_id=ModelTrainId(value=str(train_id))
            if train_id is not None
            else None,
            dataset_build_id=DatasetBuildId(value=str(dataset_build_id))
            if dataset_build_id is not None
            else None,
        )
        logged_data_pb = self._service.GetLoggedData(request=request).data
        return LoggedData(
            data=logged_data_pb.text,
            logged_data_type=LDType(logged_data_pb.type),
            tag=logged_data_pb.unique_tag,
            epoched_data=logged_data_pb.epoched_data,
        )

    def log_model_metric(
        self, train_id: UUID, tag: str, points: List[ModelMetricPoint]
    ) -> ModelMetricId:
        metric = LoggedModelMetric(
            unique_tag=tag,
            points=[
                LoggedModelMetric.ModelMetricPoint(epoch=p.epoch, value=p.value)
                for p in points
            ],
        )
        request = LogModelMetricRequest(
            model_train_id=ModelTrainId(value=str(train_id)), metric=metric
        )
        response = self._service.LogModelMetric(request=request)
        return response.model_metric_id

    def log_binary_data(
        self,
        *,
        tag: str,
        logged_data_type: "LoggedDataType.V",
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
        epoch: Optional[int] = None,
    ) -> str:
        return self._log_data(
            tag,
            logged_data_type,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
            epoch=epoch,
        ).s3_path

    def log_text_data(
        self,
        tag: str,
        data: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> None:
        self._log_data(
            tag,
            LoggedDataType.LOGGED_DATA_TYPE_TEXT,
            data=data,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
        )

    def log_markdown_data(
        self,
        tag: str,
        data: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> None:
        self._log_data(
            tag,
            LoggedDataType.LOGGED_DATA_TYPE_MARKDOWN,
            data=data,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
        )

    def log_numeric_data(
        self,
        tag: str,
        data: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> None:
        self._log_data(
            tag,
            LoggedDataType.LOGGED_DATA_TYPE_NUMBER,
            data=data,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
        )

    def log_boolean_data(
        self,
        tag: str,
        data: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> None:
        self._log_data(
            tag,
            LoggedDataType.LOGGED_DATA_TYPE_BOOLEAN,
            data=data,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
        )

    def log_table_data(
        self,
        tag: str,
        data: str,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
    ) -> None:
        self._log_data(
            tag,
            LoggedDataType.LOGGED_DATA_TYPE_TABLE,
            data=data,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
        )

    def _log_data(
        self,
        tag: str,
        type: "LoggedDataType.V",
        data: Optional[str] = None,
        train_id: Optional[UUID] = None,
        dataset_build_id: Optional[UUID] = None,
        epoch: Optional[int] = None,
    ) -> LogDataResponse:
        text = cast(str, data)
        request = LogDataRequest(
            unique_tag=tag,
            type=type,
            text=text,
            # Given that protocol buffers do not allow differentiating between default value (0) or the actual value 0
            # We use the convention that a minus value represents the absence of an epoch.
            # See https://developers.google.com/protocol-buffers/docs/proto3#default for more information.
            epoch=epoch if epoch is not None else -1,
            model_train_id=ModelTrainId(value=str(train_id))
            if train_id is not None
            else None,
            dataset_build_id=DatasetBuildId(value=str(dataset_build_id))
            if dataset_build_id is not None
            else None,
        )
        return self._service.LogData(request=request)



import uuid
from dataclasses import dataclass
from typing import Optional, Sequence, Union

import pandas as pd
from layerapi.api.ids_pb2 import ModelTrainId
from layerapi.api.value.aws_credentials_pb2 import AwsCredentials
from layerapi.api.value.s3_path_pb2 import S3Path

from layer.exceptions.exceptions import LayerClientException
from layer.flavors.base import ModelFlavor, ModelRuntimeObjects
from layer.types import ModelObject

from .asset import AssetPath, AssetType, BaseAsset


@dataclass(frozen=True)
class TrainStorageConfiguration:
    train_id: ModelTrainId
    s3_path: S3Path
    credentials: AwsCredentials


class Model(BaseAsset):
    """
    Provides access to ML models trained and stored in Layer.

    You can retrieve an instance of this object with :code:`layer.get_model()`.

    This class should not be initialized by end-users.

    .. code-block:: python

        # Fetches a specific version of this model
        layer.get_model("churn_model:1.2")

    """

    def __init__(
        self,
        asset_path: Union[str, AssetPath],
        id: Optional[uuid.UUID] = None,
        dependencies: Optional[Sequence[BaseAsset]] = None,
        version_id: Optional[uuid.UUID] = None,
        description: str = "",
        flavor: Optional[ModelFlavor] = None,
        storage_config: Optional[TrainStorageConfiguration] = None,
        model_runtime_objects: Optional[ModelRuntimeObjects] = None,
    ):
        super().__init__(
            path=asset_path,
            id=id,
            dependencies=dependencies,
        )
        self._version_id = version_id
        self._description = description
        self._flavor = flavor
        self._storage_config = storage_config
        self._model_runtime_objects = model_runtime_objects

    @property
    def asset_type(self) -> AssetType:
        return AssetType.MODEL

    @property
    def version_id(self) -> uuid.UUID:
        if self._version_id is None:
            raise LayerClientException("Model version id is not initialized")
        return self._version_id

    @property
    def flavor(self) -> ModelFlavor:
        if self._flavor is None:
            raise LayerClientException("Model flavor is not initialized")
        return self._flavor

    @property
    def storage_config(self) -> TrainStorageConfiguration:
        if self._storage_config is None:
            raise LayerClientException("Model storage config is not initialized")
        return self._storage_config

    @property
    def model_object(self) -> ModelObject:
        if self._model_runtime_objects is None:
            raise LayerClientException("Model artifact is not yet fetched from storage")
        return self._model_runtime_objects.model_object

    def get_train(self) -> ModelObject:
        """
        Returns the trained and saved model artifact. For example, a scikit-learn or PyTorch model object.

        :return: The trained model artifact.
        """
        return self.model_object

    def predict(self, input_df: pd.DataFrame) -> pd.DataFrame:
        """
        Performs prediction on the input dataframe data.
        :return: the predictions as a pd.DataFrame
        """
        if (
            self._model_runtime_objects is None
            or self._model_runtime_objects.prediction_function is None
        ):
            raise Exception("No predict function provided")
        return self._model_runtime_objects.prediction_function(input_df)

    def set_model_runtime_objects(
        self, model_runtime_objects: ModelRuntimeObjects
    ) -> "Model":
        self._model_runtime_objects = model_runtime_objects
        return self



from .context import Context  # noqa
from .contracts.datasets import Dataset  # noqa
from .contracts.logged_data import Markdown  # noqa
from .contracts.models import Model  # noqa
from .contracts.projects import Project  # noqa
from .contracts.runs import Run  # noqa
from .flavors.custom import CustomModel  # noqa
from .global_context import current_project_name  # noqa
from .logged_data.callbacks import KerasCallback, XGBoostCallback  # noqa
from .main import clear_cache  # noqa
from .main import get_dataset  # noqa
from .main import get_model  # noqa
from .main import init  # noqa
from .main import log  # noqa
from .main import login  # noqa
from .main import login_as_guest  # noqa
from .main import login_with_access_token  # noqa
from .main import login_with_api_key  # noqa
from .main import logout  # noqa
from .main import run  # noqa
from .main import show_api_key  # noqa
from .pandas_extensions import Arrays, Images, _register_type_extensions  # noqa


_register_type_extensions()



import io
import itertools
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Iterable,
    Optional,
    Sequence,
    Type,
    Union,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from pandas._typing import PositionalIndexer  # type: ignore
from pandas.core.arrays.base import ExtensionArray  # type: ignore
from pandas.core.dtypes.base import register_extension_dtype  # type: ignore
from pandas.core.dtypes.base import ExtensionDtype


if TYPE_CHECKING:
    from PIL.Image import Image


warnings.filterwarnings(
    "ignore",
    "The input object of type 'Image' is an array-like implementing one of the corresponding protocols.*",
)
warnings.filterwarnings(
    "ignore",
    "Creating an ndarray from ragged nested sequences.*",
)


_TYPE_IMAGE_NAME = "layer.image"


class _ImageType(pa.ExtensionType):
    """Arrow type extension https://arrow.apache.org/docs/python/extending_types.html#defining-extension-types-user-defined-types
    Provides addtional metadata for arrow schema about the image format.
    """

    def __init__(self) -> None:
        pa.ExtensionType.__init__(self, pa.binary(), _TYPE_IMAGE_NAME)

    def __arrow_ext_serialize__(self) -> bytes:
        return b"png"  # always store as png

    @classmethod
    def __arrow_ext_deserialize__(
        cls, storage_type: pa.DataType, serialized: bytes
    ) -> pa.ExtensionType:
        return _ImageType()


class _ImageDtype(ExtensionDtype):
    @property
    def name(self) -> str:
        return _TYPE_IMAGE_NAME

    @classmethod
    def construct_from_string(cls, string: str) -> ExtensionDtype:
        if string == _TYPE_IMAGE_NAME:
            return _ImageDtype()
        raise TypeError(f"cannot construct a '{cls.__name__}' from '{string}'")

    @classmethod
    def construct_array_type(cls) -> ExtensionArray:
        return Images

    @property
    def type(self) -> Type["Image"]:
        from PIL.Image import Image

        return Image

    def __from_arrow__(self, array: Union[pa.Array, pa.ChunkedArray]) -> "Images":
        if isinstance(array, pa.Array):
            chunks = array
        else:
            chunks = array.chunks
        images = (_load_image(binary) for chunk in chunks for binary in chunk)
        return Images(tuple(images))


def _load_image(binary: Union[pa.BinaryScalar, pa.ExtensionScalar]) -> "Image":
    storage_array = binary.value if isinstance(binary, pa.ExtensionScalar) else binary
    with io.BytesIO(storage_array.as_buffer().to_pybytes()) as buf:
        import PIL.Image

        image = PIL.Image.open(buf)
        image.load()
        return image


class Images(ExtensionArray):
    """
    Extends `pandas.ExtensionArray` to provide automatic conversion between `PIL.Image.Image` class and `pyarrow`.
    The conversion is required to store images in layer datasets.

    .. code-block:: python

        def make_image(n: int) -> Image:
            image = PIL.Image.new("RGB", (160, 40), color=(73, 109, 137))
            draw = ImageDraw.Draw(image)
            draw.text((11, 10), f"Image #{n}", fill=(255, 255, 0))
            return image

        image_1 = make_image(1)
        image_2 = make_image(2)
        image_3 = make_image(3)

        pandas.DataFrame({"images": layer.Images([image_1, image_2, image_3])})
    """

    def __init__(self, images: Sequence["Image"]):
        self._images = images
        self._dtype = _ImageDtype()

    @classmethod
    def _from_sequence(
        cls,
        scalars: Sequence[Any],
        *,
        dtype: Optional[ExtensionDtype] = None,
        copy: bool = False,
    ) -> Any:
        return cls(scalars)

    @classmethod
    def _concat_same_type(cls, to_concat: Sequence["Images"]) -> Any:
        return cls(tuple(itertools.chain(*to_concat)))

    @property
    def dtype(self) -> ExtensionDtype:
        return self._dtype

    def copy(self) -> "Images":
        return Images(tuple(image.copy() for image in self._images))

    def isna(self) -> np.ndarray:  # type: ignore
        return np.array([image is None for image in self._images], dtype=bool)

    def __len__(self) -> int:
        return len(self._images)

    def __getitem__(self, item: PositionalIndexer) -> Union["Images", "Image"]:
        if isinstance(item, int):
            # for scalar item, return a scalar value suitable for the array's type
            return self._images[item]
        if isinstance(item, np.ndarray):
            # a boolean mask, filtered to the values where item is True
            return Images(tuple(self._get_images_by_mask(item)))
        if isinstance(item, slice):
            return Images(self._images[item.start : item.stop : item.step])
        raise NotImplementedError(f"item type {type(item)}")

    def _get_images_by_mask(
        self, mask: Iterable[bool]
    ) -> Generator["Image", None, None]:
        for i, include in enumerate(mask):
            if include:
                yield self._images[i]

    def __arrow_array__(self, type: Any = None) -> pa.ExtensionArray:
        storage_array = pa.array(self._images_byte_arr(), pa.binary())
        # ignore the type and always read as _ImageType,
        # as it does not have any parameters for now
        return pa.ExtensionArray.from_storage(_ImageType(), storage_array)

    def _images_byte_arr(self) -> Generator[bytes, None, None]:
        for image in self._images:
            yield _image_bytes(image)

    def _reduce(self, name: str, *, skipna: bool = True, **kwargs: Any) -> int:
        return 0

    def __eq__(self, other: Any) -> np.ndarray:  # type: ignore
        eq_arr = np.empty(len(self), dtype=bool)
        eq_arr.fill(False)
        if not isinstance(other, Images):
            return eq_arr
        for i in range(min(len(self), len(other))):
            if np.array_equal(
                np.asarray(self._images[i]), np.asarray(other._images[i])
            ):
                eq_arr[i] = True
        return eq_arr

    def __setitem__(
        self, key: Union[int, slice, "np.ndarray[Any, Any]"], value: Any
    ) -> None:
        raise NotImplementedError()

    @property
    def nbytes(self) -> int:
        total_bytes = 0
        for buf in self._images_byte_arr():
            total_bytes += len(buf)
        return total_bytes

    def value_counts(
        values: Sequence[Any],
        sort: bool = True,
        ascending: bool = False,
        normalize: bool = False,
        bins: Any = None,
        dropna: bool = True,
    ) -> pd.Series:  # type: ignore
        # make all values unique for now
        return pd.Series(np.ones(len(values), dtype=np.int64))


def _image_bytes(image: "Image") -> bytes:
    with io.BytesIO() as buf:
        image.save(buf, format="PNG")
        return buf.getvalue()


_TYPE_NDARRAY_NAME = "layer.ndarray"


class _ArrayType(pa.ExtensionType):
    """Arrow type extension https://arrow.apache.org/docs/python/extending_types.html#defining-extension-types-user-defined-types
    Provides addtional metadata for arrow schema about the ndarray format.
    """

    def __init__(self) -> None:
        pa.ExtensionType.__init__(self, pa.binary(), _TYPE_NDARRAY_NAME)

    def __arrow_ext_serialize__(self) -> bytes:
        return b""

    @classmethod
    def __arrow_ext_deserialize__(
        cls, storage_type: pa.DataType, serialized: bytes
    ) -> pa.ExtensionType:
        return _ArrayType()


class _ArrayDtype(ExtensionDtype):
    @property
    def name(self) -> str:
        return _TYPE_NDARRAY_NAME

    @classmethod
    def construct_from_string(cls, string: str) -> ExtensionDtype:
        if string == _TYPE_NDARRAY_NAME:
            return _ArrayDtype()
        raise TypeError(f"cannot construct a '{cls.__name__}' from '{string}'")

    @classmethod
    def construct_array_type(cls) -> ExtensionArray:
        return Arrays

    @property
    def type(self) -> Type[np.ndarray]:  # type: ignore
        return np.ndarray

    def __from_arrow__(self, array: Union[pa.Array, pa.ChunkedArray]) -> "Arrays":
        if isinstance(array, pa.Array):
            chunks = array
        else:
            chunks = array.chunks
        arrays = (_load_array(binary) for chunk in chunks for binary in chunk)
        return Arrays(tuple(arrays))


def _load_array(binary: Union[pa.BinaryScalar, pa.ExtensionScalar]) -> np.ndarray:  # type: ignore
    storage_array = binary.value if isinstance(binary, pa.ExtensionScalar) else binary
    with io.BytesIO(storage_array.as_buffer().to_pybytes()) as buf:
        return np.load(buf, allow_pickle=False, fix_imports=False)  # type: ignore


class Arrays(ExtensionArray):
    """
    Extends `pandas.ExtensionArray` to provide automatic conversion between `numpy.ndarray` class and `pyarrow`.
    The conversion is required to store multi-dimensional arrays in layer datasets.

    .. code-block:: python

        pd.DataFrame(
        {
            "array": layer.Arrays(
                (
                    np.array([[1, 2, 3], [4, 5, 6]], dtype=np.int64),
                    np.array([[7, 8, 9], [10, 11, 12]], dtype=np.int64),
                    np.array([[13, 14, 15], [16, 17, 18]], dtype=np.int64),
                )
            )
        })
    """

    def __init__(self, arrays: Sequence[np.ndarray]):  # type: ignore
        self._arrays = arrays
        self._dtype = _ArrayDtype()

    @classmethod
    def _from_sequence(
        cls,
        scalars: Sequence[Any],
        *,
        dtype: Optional[ExtensionDtype] = None,
        copy: bool = False,
    ) -> "Arrays":
        return Arrays(scalars)

    @property
    def dtype(self) -> ExtensionDtype:
        return self._dtype

    def copy(self) -> "Arrays":
        return Arrays(tuple(arr.copy() for arr in self._arrays))

    def isna(self) -> np.ndarray:  # type: ignore
        return np.array([arr is None for arr in self._arrays], dtype=bool)

    def __len__(self) -> int:
        return len(self._arrays)

    def __getitem__(self, item: PositionalIndexer) -> Union["Arrays", np.ndarray]:  # type: ignore
        if isinstance(item, int):
            # for scalar item, return a scalar value suitable for the array's type
            return self._arrays[item]
        if isinstance(item, np.ndarray):
            # a boolean mask, filtered to the values where item is True
            return Arrays(tuple(self._get_arrays_by_mask(item)))
        if isinstance(item, slice):
            return Arrays(self._arrays[item.start : item.stop : item.step])
        raise NotImplementedError(f"item type {type(item)}")

    def _get_arrays_by_mask(
        self, mask: Iterable[bool]
    ) -> Generator[np.ndarray, None, None]:  # type: ignore
        for i, include in enumerate(mask):
            if include:
                yield self._arrays[i]

    def __arrow_array__(self, type: Any = None) -> pa.ExtensionArray:
        storage_array = pa.array(
            (_array_bytes(arr) for arr in self._arrays), pa.binary()
        )
        return pa.ExtensionArray.from_storage(_ArrayType(), storage_array)

    def __setitem__(
        self, key: Union[int, slice, "np.ndarray[Any, Any]"], value: Any
    ) -> None:
        raise NotImplementedError()

    @property
    def nbytes(self) -> int:
        total_bytes = 0
        for arr in self._arrays:
            total_bytes += arr.nbytes
        return total_bytes

    def value_counts(
        values: Sequence[Any],
        sort: bool = True,
        ascending: bool = False,
        normalize: bool = False,
        bins: Any = None,
        dropna: bool = True,
    ) -> pd.Series:  # type: ignore
        # make all values unique for now
        return pd.Series(np.ones(len(values), dtype=np.int64))


def _array_bytes(arr: np.ndarray) -> bytes:  # type: ignore
    with io.BytesIO() as buf:
        np.save(buf, arr, allow_pickle=False, fix_imports=False)  # type: ignore
        return buf.getvalue()


def _register_type_extensions() -> None:
    # register arrow extension types
    pa.register_extension_type(_ImageType())
    pa.register_extension_type(_ArrayType())

    # register pandas extension types
    register_extension_dtype(_ImageDtype)
    register_extension_dtype(_ArrayDtype)


def _infer_custom_types(data: pd.DataFrame) -> pd.DataFrame:
    if len(data) == 0:
        return data
    first_row = data.head(n=1)
    try:
        from PIL.Image import Image

        for col in first_row:
            value = data[col][0]
            if isinstance(value, Image):
                data[col] = Images(data[col])
            elif isinstance(value, np.ndarray) and value.ndim > 1:
                data[col] = Arrays(data[col])
    except ImportError:
        pass
    return data



from dataclasses import dataclass


@dataclass(frozen=True)
class ProjectFullName:
    """
    Represents the project name and its owning account name
    """

    account_name: str
    project_name: str

    @property
    def path(self) -> str:
        return f"{self.account_name}/{self.project_name}"



from pathlib import Path

import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class CatBoostModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of CatBoost Models."""

    MODULE_KEYWORD = "catboost"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_CATBOOST

    def save_model_to_directory(
        self, model_object: ModelObject, directory: Path
    ) -> None:
        import mlflow.catboost

        mlflow.catboost.save_model(model_object, path=directory.as_posix())

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        import mlflow.catboost

        model = mlflow.catboost.load_model(directory.as_uri())
        return ModelRuntimeObjects(
            model, lambda input_df: self.__predict(model, input_df)
        )

    @staticmethod
    def __predict(model: ModelObject, input_df: pd.DataFrame) -> pd.DataFrame:
        prediction_np_array = model.predict(input_df)  # type: ignore
        return pd.DataFrame(prediction_np_array)



from typing import Any, List, Optional

from layer.contracts.assertions import Assertion
from layer.contracts.assets import AssetPath, AssetType
from layer.contracts.fabrics import Fabric
from layer.contracts.runs import ResourcePath
from layer.exceptions.exceptions import ConfigError, LayerClientException
from layer.global_context import (
    default_fabric,
    get_pip_packages,
    get_pip_requirements_file,
)


def _resolve_settings(
    override: Optional[Any], default: Optional[Any], fallback: Optional[Any]
) -> Any:
    setting = override if override else default
    return setting if setting else fallback


class LayerSettings:
    _asset_type: Optional[AssetType] = None
    _name: Optional[str] = None
    _fabric: Optional[Fabric] = None
    _pip_requirements_file: Optional[str] = None
    _pip_packages: Optional[List[str]] = None
    _resource_paths: Optional[List[ResourcePath]] = None
    _dependencies: Optional[List[AssetPath]] = None
    _assertions: Optional[List[Assertion]] = None

    def get_asset_type(self) -> AssetType:
        if self._asset_type is None:
            raise LayerClientException("Asset type cannot be empty")
        return self._asset_type

    def get_asset_name(self) -> str:
        if self._name is None:
            raise LayerClientException("Asset name cannot be empty")
        return self._name

    def get_fabric(self) -> Fabric:
        return _resolve_settings(self._fabric, default_fabric(), Fabric.default())

    def get_pip_requirements_file(self) -> str:
        return _resolve_settings(
            self._pip_requirements_file, get_pip_requirements_file(), ""
        )

    def get_pip_packages(self) -> List[str]:
        return _resolve_settings(self._pip_packages, get_pip_packages(), [])

    def get_resource_paths(self) -> List[ResourcePath]:
        return self._resource_paths or []

    def get_dependencies(self) -> List[AssetPath]:
        return self._dependencies or []

    def get_assertions(self) -> List[Assertion]:
        return self._assertions or []

    def set_asset_type(self, asset_type: AssetType) -> None:
        self._asset_type = asset_type

    def set_asset_name(self, name: str) -> None:
        self._name = name

    def set_fabric(self, f: str) -> None:
        if Fabric.has_member_key(f):
            self._fabric = Fabric(f)
            return
        raise ValueError(
            'Fabric setting "{}" is not valid. You can check valid values in Fabric enum definition.'.format(
                f
            )
        )

    def set_pip_requirements_file(self, file: Optional[str]) -> None:
        self._pip_requirements_file = file

    def set_pip_packages(self, packages: Optional[List[str]]) -> None:
        self._pip_packages = packages

    def set_resource_paths(self, paths: Optional[List[ResourcePath]]) -> None:
        self._resource_paths = paths

    def set_dependencies(self, dependencies: List[AssetPath]) -> None:
        self._dependencies = dependencies

    def append_assertion(self, assertion: Assertion) -> None:
        self._assertions = self._assertions or []
        self._assertions.append(assertion)

    def validate(self) -> None:
        if self._asset_type is None:
            raise ConfigError(
                'Either @dataset(name="...") or @model(name="...") top level decorator '
                "is required for each function. Add @dataset or @model decorator on top of existing "
                "decorators to run functions in Layer."
            )
        if self._name is None or self._name == "":
            raise ConfigError(
                "Your @dataset and @model must be named. Pass an asset name as a first argument to your decorators."
            )
        fabric = self.get_fabric()
        if (
            fabric is not None
            and fabric.is_gpu()
            and self.get_asset_type() is AssetType.DATASET
        ):
            raise ConfigError(
                "GPU fabrics can only be used for model training. Use a different fabric for your dataset build."
            )



import importlib.util
from typing import Any, Optional, cast

import pandas as pd
import pkg_resources
from packaging import version

import layer


# We need keras and xgboost libraries to be able to define XGBoostCallback and KerasCallback convenience classes.
# Since these are quite large libraries, instead of directly bundling them into our Python package, we use a hacky
# solution to have a "skeleton" base class if the Python environment doesn't already have these libraries installed.
# Given that one would almost certainly have these libraries installed if they decide to use the convenience classes,
# this is a good tradeoff.
# XGBoost implemented the callback interface at version 1.3.0, make sure we have a version newer than that.
if importlib.util.find_spec("xgboost") is None or version.parse(
    pkg_resources.get_distribution("xgboost").version
) < version.parse("1.3.0"):

    class XGBoostTrainingCallback:
        def before_training(self, model: Any) -> Any:
            pass

        def after_training(self, model: Any) -> Any:
            pass

        def after_iteration(self, model: Any, epoch: int, evals_log: dict) -> bool:  # type: ignore
            pass

else:
    import xgboost as xgb

    class XGBoostTrainingCallback(xgb.callback.TrainingCallback):  # type: ignore
        pass


if importlib.util.find_spec("keras") is None:

    class KerasTrainingCallback:
        def on_epoch_end(self, epoch: int, logs: Optional[dict] = None) -> None:  # type: ignore
            pass

else:
    import keras  # type: ignore

    class KerasTrainingCallback(keras.callbacks.Callback):  # type: ignore
        pass


class XGBoostCallback(XGBoostTrainingCallback):
    """
    A default implementation for `callbacks` parameter in various XGBoost methods that uses `layer.log`.

    This class is purely for convenience and one can create their own implementations with different overrides.
    For more information, please consult to XGBoost documentation at https://xgboost.readthedocs.io/en/stable/python/callbacks.html
    """

    def __init__(self, importance_type: str = "gain") -> None:
        super().__init__()
        self.importance_type = importance_type

    def before_training(self, model: Any) -> Any:
        return model

    def after_training(self, model: Any) -> Any:
        if model.attr("best_score") is not None:
            print(
                {
                    "best_score": float(cast(str, model.attr("best_score"))),
                    "best_iteration": int(cast(str, model.attr("best_iteration"))),
                }
            )

        features = model.get_score(importance_type=self.importance_type)
        importance = [[feature, features[feature]] for feature in features]

        df = pd.DataFrame(data=importance, columns=["feature", "importance"])
        layer.log({"Feature Importance Table": df})
        layer.log(
            {"Feature Importance": df.plot.bar(x="feature", y="importance", rot=0)}  # type: ignore
        )

        return model

    def after_iteration(self, model: Any, epoch: int, evals_log: dict) -> bool:  # type: ignore
        if evals_log:
            layer.log(evals_log, epoch)
        return False


class KerasCallback(KerasTrainingCallback):
    """
    A default implementation for `callbacks` parameter in various Keras methods that uses `layer.log`.

    This class is purely for convenience and one can create their own implementations with different overrides.
    For more information, please consult to Keras documentation at https://keras.io/api/callbacks/
    """

    def on_epoch_end(self, epoch: int, logs: Optional[dict] = None) -> None:  # type: ignore
        if logs:
            layer.log(logs, epoch)



import uuid
from collections import defaultdict, deque
from dataclasses import dataclass
from typing import TYPE_CHECKING, DefaultDict, List, Optional, Sequence, Set

from layerapi.api.entity.operations_pb2 import (
    DatasetBuildOperation,
    ExecutionPlan,
    ModelTrainOperation,
    Operation,
    ParallelOperation,
    SequentialOperation,
)
from layerapi.api.ids_pb2 import ModelVersionId

from layer.contracts.assets import AssetPath, AssetType
from layer.contracts.runs import FunctionDefinition, Run
from layer.exceptions.exceptions import (
    LayerClientException,
    ProjectCircularDependenciesException,
    ProjectDependencyNotFoundException,
)


if TYPE_CHECKING:
    from networkx import DiGraph  # type: ignore


@dataclass(frozen=True)
class PlanNode:
    path: AssetPath
    dependencies: List[AssetPath]
    name: str
    id: Optional[uuid.UUID]


def build_execution_plan(run: Run) -> ExecutionPlan:
    graph = _build_directed_acyclic_graph(run.definitions)
    plan = _topological_sort_grouping(graph)
    operations = []
    for _level, ops in plan.items():
        if len(ops) > 1:
            dataset_build_operations = []
            model_train_operations = []
            for operation in ops:
                dependencies = [d.path() for d in operation.dependencies]
                if operation.path.asset_type == AssetType.MODEL:
                    model_train_operation = ModelTrainOperation(
                        model_version_id=ModelVersionId(value=str(operation.id)),
                        dependency=dependencies,
                    )
                    model_train_operations.append(model_train_operation)
                elif operation.path.asset_type == AssetType.DATASET:
                    dataset_build_operation = DatasetBuildOperation(
                        dataset_name=operation.name,
                        dependency=dependencies,
                    )
                    dataset_build_operations.append(dataset_build_operation)
                else:
                    raise LayerClientException(f"Unknown operation type. {operation}")
            operations.append(
                Operation(
                    parallel=ParallelOperation(
                        dataset_build=dataset_build_operations,
                        model_train=model_train_operations,
                    )
                )
            )
        else:
            (operation,) = ops
            dependencies = [d.path() for d in operation.dependencies]
            if operation.path.asset_type == AssetType.MODEL:
                model_train_operation = ModelTrainOperation(
                    model_version_id=ModelVersionId(value=str(operation.id)),
                    dependency=dependencies,
                )
                operations.append(
                    Operation(
                        sequential=SequentialOperation(
                            model_train=model_train_operation
                        )
                    )
                )
            elif operation.path.asset_type == AssetType.DATASET:
                dataset_build_operation = DatasetBuildOperation(
                    dataset_name=operation.name,
                    dependency=dependencies,
                )
                operations.append(
                    Operation(
                        sequential=SequentialOperation(
                            dataset_build=dataset_build_operation
                        )
                    )
                )
            else:
                raise LayerClientException(f"Unknown operation type. {operation}")
    execution_plan = ExecutionPlan(operations=operations)
    return execution_plan


def check_asset_dependencies(definitions: Sequence[FunctionDefinition]) -> None:
    _build_directed_acyclic_graph(definitions)


def drop_independent_entities(
    definitions: Sequence[FunctionDefinition],
    target_path: AssetPath,
    *,
    keep_dependencies: bool = True,
) -> Sequence[FunctionDefinition]:
    from networkx import NodeNotFound, shortest_path

    target_asset_id = _get_asset_id(target_path)
    if keep_dependencies:
        graph = _build_directed_acyclic_graph(definitions)
        try:
            asset_ids: Set[str] = set.union(
                set(), *shortest_path(graph, target=target_asset_id).values()
            )
        except NodeNotFound:
            raise _create_not_found_exception(target_path)
    else:
        asset_ids = {target_asset_id}
    definitions = [
        f if keep_dependencies else f.drop_dependencies()
        for f in definitions
        if _get_asset_id(f.asset_path) in asset_ids
    ]
    return definitions


def _build_graph(definitions: Sequence[FunctionDefinition]) -> "DiGraph":
    from networkx import DiGraph

    graph = DiGraph()

    for func in definitions:
        _add_function_to_graph(graph, func)

    for func in definitions:
        asset_id = _get_asset_id(func.asset_path)
        for dependency_path in func.asset_dependencies:
            dependency_asset_id = _get_asset_id(dependency_path)
            # we add connections only to other entities to build
            if dependency_asset_id in graph.nodes:
                graph.add_edge(dependency_asset_id, asset_id)

    return graph


def _build_directed_acyclic_graph(
    definitions: Sequence[FunctionDefinition],
) -> "DiGraph":
    from networkx import is_directed_acyclic_graph

    graph = _build_graph(definitions)

    if not is_directed_acyclic_graph(graph):
        cycles: List[List[AssetPath]] = _find_cycles(graph.reverse())
        stringified_paths = [_stringify_asset_cycle(cycle) for cycle in cycles]
        stringified_paths.sort()  # Ensure stability across different runs
        raise ProjectCircularDependenciesException(stringified_paths)
    return graph


def _add_function_to_graph(graph: "DiGraph", func: FunctionDefinition) -> None:
    graph.add_node(
        _get_asset_id(func.asset_path),
        node=PlanNode(
            path=func.asset_path,
            name=func.asset_name,
            id=func.version_id,
            dependencies=func.asset_dependencies,
        ),
    )


def _find_cycles(graph: "DiGraph") -> List[List[AssetPath]]:
    from networkx import get_node_attributes, simple_cycles

    cycle_paths: List[List[AssetPath]] = []
    entities_map = get_node_attributes(graph, "node")
    for cycle in simple_cycles(graph):
        cycle_path: List[AssetPath] = [entities_map[node] for node in cycle]
        cycle_paths.append(cycle_path)
    return cycle_paths


def _topological_sort_grouping(
    graph: "DiGraph",
) -> DefaultDict[int, List[PlanNode]]:
    _graph = graph.copy()
    res = defaultdict(list)
    level = 0
    while _graph:
        zero_in_degree = [
            vertex for vertex, degree in _graph.in_degree() if degree == 0
        ]
        res[level] = [_graph.nodes[vertex]["node"] for vertex in zero_in_degree]
        _graph.remove_nodes_from(zero_in_degree)
        level = level + 1
    return res


def _stringify_asset_cycle(asset_cycle_path: List[AssetPath]) -> str:
    def rotate_list(asset_cycle_path: List[str]) -> List[str]:
        smallest_idx = 0
        for i in range(1, len(asset_cycle_path)):
            if asset_cycle_path[i] < asset_cycle_path[smallest_idx]:
                smallest_idx = i
        rotated = deque(asset_cycle_path)
        rotated.rotate(-smallest_idx)
        return list(rotated)

    stringified = rotate_list(
        [str(asset) for asset in asset_cycle_path]
    )  # Ensure stability within a cycle path across runs
    stringified.append(stringified[0])  # Add dependency to first node in cycle
    return " -> ".join(stringified)


def _get_asset_id(path: AssetPath) -> str:
    if path.project_name is None:
        raise LayerClientException("Cannot plan execution with missing project name")
    return path.path()


def _create_not_found_exception(path: AssetPath) -> Exception:
    return ProjectDependencyNotFoundException(
        f"{path.path()} not found. Declare the dependency in your project",
    )



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from logging import Logger
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, List, Optional
from uuid import UUID

from layerapi.api.entity.model_train_status_pb2 import ModelTrainStatus

from layer import Context
from layer.clients.layer import LayerClient
from layer.contracts.assertions import Assertion
from layer.exceptions.exception_handler import exception_handler
from layer.exceptions.exceptions import LayerFailedAssertionsException
from layer.global_context import (
    current_project_full_name,
    reset_active_context,
    set_active_context,
)
from layer.resource_manager import ResourceManager
from layer.tracker.progress_tracker import RunProgressTracker
from layer.training.train import Train

from .common import import_function, update_train_status
from .model_train_failure_reporter import ModelTrainFailureReporter


@dataclass
class TrainContextDataclassMixin:
    model_name: str
    model_version: str
    train_id: UUID
    source_entrypoint: str
    source_folder: Path
    logger: Logger
    train_index: Optional[str] = None


class TrainContext(ABC, TrainContextDataclassMixin):
    def init_or_save_context(self, context: Context) -> None:
        set_active_context(context)

    @abstractmethod
    def __enter__(self) -> None:
        pass

    @abstractmethod
    def get_working_directory(self) -> Path:
        pass

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        reset_active_context()


@dataclass
class LocalTrainContext(TrainContext):
    initial_cwd: Optional[str] = None

    def __enter__(self) -> None:
        super().__enter__()
        self.initial_cwd = os.getcwd()

    def get_working_directory(self) -> Path:
        assert self.initial_cwd
        return Path(self.initial_cwd)

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        assert self.initial_cwd
        os.chdir(
            self.initial_cwd
        )  # Important for local execution to have no such side effect
        return None


@dataclass(frozen=True)
class ModelTrainer:
    client: LayerClient
    train_context: TrainContext
    logger: Logger
    failure_reporter: ModelTrainFailureReporter
    tracker: RunProgressTracker

    def train(self) -> Any:
        self.tracker.mark_model_training(
            self.train_context.model_name,
            version=self.train_context.model_version,
            train_idx=self.train_context.train_index,
        )
        try:
            with self.train_context:
                return self._train(callback=self.failure_reporter.report_failure)
        except Exception:
            self.logger.error(
                f"Error performing model training with id: {self.train_context.train_id}",
                exc_info=True,
            )
            import sys
            import traceback

            traceback.print_exc(file=sys.stdout)
            sys.exit(1)

    def _run_assertions(self, model: Any, assertions: List[Assertion]) -> None:
        failed_assertions = []
        self.tracker.mark_model_running_assertions(self.train_context.model_name)
        for assertion in reversed(assertions):
            try:
                self.tracker.mark_model_running_assertion(
                    self.train_context.model_name, assertion
                )
                assertion.function(model)
            except Exception:
                failed_assertions.append(assertion)
        if len(failed_assertions) > 0:
            self.tracker.mark_model_failed_assertions(
                self.train_context.model_name, failed_assertions
            )
            raise LayerFailedAssertionsException(failed_assertions)
        else:
            self.tracker.mark_model_completed_assertions(self.train_context.model_name)

    @exception_handler(stage="Training run")
    def _train(
        self, callback: Optional[Callable[[str, Exception], None]] = None
    ) -> Any:
        self.logger.info(
            f"Importing user code({self.train_context.source_entrypoint}) from {self.train_context.source_folder}"
        )
        train_model_func = import_function(
            self.train_context.source_folder,
            self.train_context.source_entrypoint,
            "train_model",
        )
        self.logger.info("train_model_func function imported successfully")
        project_full_name = current_project_full_name()
        if not project_full_name:
            raise Exception("Internal Error: missing current project full name")
        with Context() as context:
            with Train(
                layer_client=self.client,
                name=self.train_context.model_name,
                project_full_name=project_full_name,
                version=self.train_context.model_version,
                train_id=self.train_context.train_id,
            ) as train:
                context.with_train(train)
                context.with_tracker(self.tracker)
                context.with_asset_name(self.train_context.model_name)
                self.train_context.init_or_save_context(context)
                update_train_status(
                    self.client.model_catalog,
                    self.train_context.train_id,
                    ModelTrainStatus.TRAIN_STATUS_FETCHING_FEATURES,
                    self.logger,
                )

                update_train_status(
                    self.client.model_catalog,
                    self.train_context.train_id,
                    ModelTrainStatus.TRAIN_STATUS_IN_PROGRESS,
                    self.logger,
                )
                self.logger.info("Executing the train_model_func")
                work_dir = self.train_context.get_working_directory()
                os.chdir(work_dir)
                self.logger.info("Downloading resources")
                ResourceManager(self.client).wait_resource_download(
                    project_full_name,
                    train_model_func.__name__,
                    target_dir=str(work_dir),
                )
                model = train_model_func()
                self.tracker.mark_model_trained(
                    self.train_context.model_name,
                )
                self.logger.info("Executed train_model_func successfully")
                self._run_assertions(
                    model,
                    train_model_func.layer.get_assertions(),  # type: ignore
                )
                self.tracker.mark_model_saving(self.train_context.model_name)
                self.logger.info(f"Saving model artifact {model} to model registry")
                train.save_model(model, tracker=self.tracker)
                update_train_status(
                    self.client.model_catalog,
                    self.train_context.train_id,
                    ModelTrainStatus.TRAIN_STATUS_SUCCESSFUL,
                    self.logger,
                )
                self.logger.info(
                    f"Saved model artifact {model} to model registry successfully"
                )
                self.tracker.mark_model_saved(self.train_context.model_name)
                return model



import asyncio
import concurrent.futures
from typing import Any, Awaitable


def asyncio_run_in_thread(coro: Awaitable[Any]) -> Any:
    def _run() -> None:
        return asyncio.run(coro)

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        for future in concurrent.futures.as_completed([executor.submit(_run)]):
            if future.exception():
                raise future.exception()  # type: ignore
            return future.result()



import logging
import warnings
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Union,
)
from uuid import UUID

from yarl import URL

from layer.cache.cache import Cache
from layer.cache.utils import is_cached
from layer.clients.layer import LayerClient
from layer.config import DEFAULT_PATH, DEFAULT_URL, ConfigManager
from layer.config.config import Config
from layer.context import Context
from layer.contracts.assets import AssetPath, AssetType
from layer.contracts.datasets import Dataset
from layer.contracts.fabrics import Fabric
from layer.contracts.models import Model
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.projects import Project
from layer.contracts.runs import Run
from layer.contracts.tracker import ResourceTransferState
from layer.exceptions.exceptions import (
    ConfigError,
    ProjectInitializationException,
    UserConfigurationError,
    UserNotLoggedInException,
    UserWithoutAccountError,
)
from layer.global_context import (
    current_account_name,
    current_project_full_name,
    current_project_name,
    get_active_context,
    has_shown_python_version_message,
    has_shown_update_message,
    reset_active_context,
    reset_to,
    set_active_context,
    set_has_shown_python_version_message,
    set_has_shown_update_message,
)
from layer.logged_data.log_data_runner import LogDataRunner
from layer.projects.init_project_runner import InitProjectRunner
from layer.projects.project_runner import ProjectRunner
from layer.projects.utils import get_current_project_name
from layer.settings import LayerSettings
from layer.tracker.progress_tracker import RunProgressTracker
from layer.training.train import Train
from layer.utils.async_utils import asyncio_run_in_thread


if TYPE_CHECKING:
    import matplotlib.axes._subplots  # type: ignore
    import matplotlib.figure  # type: ignore
    import pandas
    import PIL.Image


logger = logging.getLogger(__name__)

warnings.filterwarnings(
    "ignore",
    "`should_run_async` will not call `transform_cell` automatically in the future.",
    DeprecationWarning,
)


def login(url: Union[URL, str] = DEFAULT_URL) -> None:
    """
    :param url: Not used.
    :raise UserNotLoggedInException: If the user is not logged in.
    :raise UserConfigurationError: If Layer user is not configured correctly.

    Logs user in to Layer. You might be prompted to enter an access token from a web page that is provided.

    .. code-block:: python

        layer.login()

    """
    _check_latest_version()

    async def _login(manager: ConfigManager, login_url: URL) -> None:
        await manager.login_headless(login_url)

    _refresh_and_login_if_needed(url, _login)


def login_with_access_token(
    access_token: str, url: Union[URL, str] = DEFAULT_URL
) -> None:
    """
    :param access_token: A valid access token retrieved from Layer.
    :param url: Not used.
    :raise UserNotLoggedInException: If the user is not logged in.
    :raise UserConfigurationError: If Layer user is not configured correctly.

    Log in with an access token. You might be prompted to enter an access token from a web page that is provided.

    .. code-block:: python

        layer.login_with_access_token(TOKEN)
    """

    async def _login(manager: ConfigManager, login_url: URL) -> None:
        await manager.login_with_access_token(login_url, access_token)

    _refresh_and_login_if_needed(url, _login)


def _refresh_and_login_if_needed(
    url: Union[URL, str], login_func: Callable[[ConfigManager, URL], Awaitable[None]]
) -> None:
    async def _login(login_url: URL) -> None:
        manager = ConfigManager(DEFAULT_PATH)
        try:
            config = await manager.refresh()
            if not config.is_guest and config.url == login_url:
                # no need to re-login
                return
        except (UserNotLoggedInException, UserConfigurationError):
            pass

        try:
            await login_func(manager, login_url)
        except UserWithoutAccountError as ex:
            raise SystemExit(ex)

    if isinstance(url, str):
        url = URL(url)

    asyncio_run_in_thread(_login(url))


def login_as_guest(url: Union[URL, str] = DEFAULT_URL) -> None:
    """
    :param url: (optional) The target platform where the user wants to log in.

    Logs user in to Layer as a guest user. Guest users can view public projects, but they cannot add, edit, or delete entities.

    .. code-block:: python

        layer.login_as_guest() # Uses default target platform URL

    """

    async def _login_as_guest(url: URL) -> None:
        manager = ConfigManager(DEFAULT_PATH)
        await manager.login_as_guest(url)

    if isinstance(url, str):
        url = URL(url)

    asyncio_run_in_thread(_login_as_guest(url))


def login_with_api_key(api_key: str, url: Union[URL, str] = DEFAULT_URL) -> None:
    """
    :param access_token: A valid API key retrieved from Layer UI.
    :param url: Not used.
    :raise AuthException: If the API key is invalid.
    :raise UserConfigurationError: If Layer user is not configured correctly.

    Log in with an API key.

    .. code-block:: python

        layer.login_with_api_key(API_KEY)
    """

    async def _login(url: URL) -> None:
        manager = ConfigManager(DEFAULT_PATH)
        await manager.login_with_api_key(url, api_key)

    if isinstance(url, str):
        url = URL(url)

    asyncio_run_in_thread(_login(url))


def logout() -> None:
    """
    Log out of Layer.

    .. code-block:: python

        layer.logout()

    """
    asyncio_run_in_thread(ConfigManager(DEFAULT_PATH).logout())


def show_api_key() -> None:
    config = ConfigManager(DEFAULT_PATH).load()
    print(config.credentials.refresh_token)


def get_dataset(name: str, no_cache: bool = False) -> Dataset:
    """
    :param name: Name or path of the dataset.
    :param no_cache: if True, force dataset fetch from the remote location.
    :return: A dataset object defined in a Layer project.

    Retrieves a Layer dataset object from the Discover > Datasets tab.

    Guest users can use this function to access public datasets without logging in to Layer.

    By default, this function caches dataset contents locally when possible.

    .. code-block:: python

        # The simplest application of this method just returns a list of dataset information.
        layer.get_dataset("titanic")

        # This example gets the titanic dataset, turns it into a Pandas dataframe, and then displays it.
        dataset = layer.get_dataset("titanic")
        df = dataset.to_pandas()
        df.head()

        # You can also get datasets from other projects you have access to, including public projects.
        layer.get_dataset("the-project/datasets/titanic")
    """
    config: Config = asyncio_run_in_thread(ConfigManager().refresh(allow_guest=True))
    asset_path = AssetPath.parse(name, expected_asset_type=AssetType.DATASET)
    asset_path = _ensure_asset_path_is_absolute(asset_path)

    def fetch_dataset() -> "pandas.DataFrame":
        context = get_active_context()
        with LayerClient(config.client, logger).init() as client:
            within_run = (
                True if context else False
            )  # if layer.get_dataset is called within an @dataset decorated func or not
            callback = lambda: client.data_catalog.fetch_dataset(  # noqa: E731
                asset_path, no_cache=no_cache
            )
            if not within_run:
                try:
                    with Context() as context:
                        set_active_context(context)
                        context.with_asset_name(asset_path.asset_name)
                        context.with_asset_type(AssetType.DATASET)
                        tracker = RunProgressTracker(
                            url=config.url,
                            account_name=asset_path.must_org_name(),
                            project_name=asset_path.must_project_name(),
                        )
                        context.with_tracker(tracker)
                        with tracker.track():
                            dataset = _ui_progress_with_tracker(
                                callback,
                                asset_path.asset_name,
                                False,  # Datasets are fetched per partition, no good way to show caching per partition
                                within_run,
                                context,
                                AssetType.DATASET,
                            )
                finally:
                    reset_active_context()  # Reset only if outside layer func, as the layer func logic will reset it
            else:
                assert context
                dataset = _ui_progress_with_tracker(
                    callback,
                    asset_path.asset_name,
                    False,  # Datasets are fetched per partition, no good way to show caching per partition
                    within_run,
                    context,
                    AssetType.DATASET,
                )

            return dataset

    return Dataset(
        asset_path=asset_path,
        _pandas_df_factory=fetch_dataset,
    )


def get_model(name: str, no_cache: bool = False) -> Model:
    """
    :param name: Name or path of the model. You can pass additional parameters in the name to retrieve a specific version of the model with format: ``model_name:major_version.minor_version``
    :param no_cache: if True, force model fetch from the remote location.
    :return: The model object.

    Retrieves a Layer model object by its name.

    Guest users can use this function to access public models without logging in to Layer.

    By default, this function caches models locally when possible.

    .. code-block:: python

        # Loads the default version of the model.
        layer.get_model("churn_model")

        # Loads the latest train of version 2.
        layer.get_model("churn_model:2")

        # Loads a specific train of the model version 2.
        layer.get_model("churn_model:2.12")

        # Load a model from a project you aren't logged in to.
        layer.get_model("the-project/models/churn_model")
    """
    config = asyncio_run_in_thread(ConfigManager().refresh(allow_guest=True))
    asset_path = AssetPath.parse(name, expected_asset_type=AssetType.MODEL)
    asset_path = _ensure_asset_path_is_absolute(asset_path)
    context = get_active_context()

    with LayerClient(config.client, logger).init() as client:
        within_run = (
            True if context else False
        )  # if layer.get_model is called within a @model decorated func of not
        model = client.model_catalog.load_model_by_path(path=asset_path.path())
        from_cache = not no_cache and is_cached(model)
        state = ResourceTransferState(model.name)

        def callback() -> Model:
            return _load_model_runtime_objects(client, model, state, no_cache)

        if not within_run:
            try:
                with Context() as context:
                    set_active_context(context)
                    context.with_asset_name(asset_path.asset_name)
                    context.with_asset_type(AssetType.MODEL)
                    tracker = RunProgressTracker(
                        url=config.url,
                        account_name=asset_path.must_org_name(),
                        project_name=asset_path.must_project_name(),
                    )
                    context.with_tracker(tracker)
                    with tracker.track():
                        model = _ui_progress_with_tracker(
                            callback,
                            asset_path.asset_name,
                            from_cache,
                            within_run,
                            context,
                            AssetType.MODEL,
                            state,
                        )
            finally:
                reset_active_context()  # Reset only if outside layer func, as the layer func logic will reset it
        else:
            assert context
            model = _ui_progress_with_tracker(
                callback,
                asset_path.asset_name,
                from_cache,
                within_run,
                context,
                AssetType.MODEL,
                state,
            )

        return model


def _load_model_runtime_objects(
    client: LayerClient,
    model: Model,
    state: ResourceTransferState,
    no_cache: bool,
) -> Model:
    model_runtime_objects = client.model_catalog.load_model_runtime_objects(
        model,
        state=state,
        no_cache=no_cache,
    )
    model.set_model_runtime_objects(model_runtime_objects)
    return model


def _ui_progress_with_tracker(
    callback: Callable[[], Any],
    getting_asset_name: str,
    from_cache: bool,
    within_run: bool,
    context: Context,
    getting_asset_type: AssetType,
    state: Optional[ResourceTransferState] = None,
) -> Any:
    asset_name = context.asset_name()
    assert asset_name
    tracker = context.tracker()
    assert tracker
    asset_type = context.asset_type()
    assert asset_type
    if asset_type == AssetType.MODEL:
        if getting_asset_type == AssetType.DATASET:
            tracker.mark_model_getting_dataset(
                asset_name, getting_asset_name, from_cache
            )
        elif getting_asset_type == AssetType.MODEL:
            tracker.mark_model_getting_model(
                asset_name, getting_asset_name, state, from_cache
            )
    elif asset_type == AssetType.DATASET:
        if getting_asset_type == AssetType.DATASET:
            tracker.mark_dataset_getting_dataset(
                asset_name, getting_asset_name, from_cache
            )
        elif getting_asset_type == AssetType.MODEL:
            tracker.mark_dataset_getting_model(
                asset_name, getting_asset_name, state, from_cache
            )
    result = callback()
    if within_run:
        if asset_type == AssetType.MODEL:
            tracker.mark_model_training(asset_name)
        elif asset_type == AssetType.DATASET:
            tracker.mark_dataset_building(asset_name)
    elif asset_type == AssetType.MODEL:
        tracker.mark_model_loaded(asset_name)
    elif asset_type == AssetType.DATASET:
        tracker.mark_dataset_loaded(asset_name)
    return result


def _ensure_asset_path_is_absolute(
    path: AssetPath,
) -> AssetPath:
    if not path.is_relative():
        return path
    project_name = path.project_name if path.has_project() else current_project_name()
    account_name = (
        path.org_name if path.org_name is not None else current_account_name()
    )

    if not project_name or not account_name:
        raise ProjectInitializationException(
            "Please specify the project full name globally with layer.init('account-name/project-name')"
            "or have it in the asset full name like 'the-account/the-project/models/the-model-name'"
        )

    path = path.with_project_full_name(
        ProjectFullName(account_name=account_name, project_name=project_name)
    )
    return path


@contextmanager
def start_train(
    name: str, version: str, train_id: Optional[UUID] = None
) -> Iterator[Train]:
    """
    :param name: name of the model
    :param version: version number of the model
    :param train_id: id to start the training with

    Initiates a model training, generating and retaining model metadata on the backend

    Example usage:
    import layer

    with layer.start_train(name="model_name", version=2) as train:
        train.register_input(x_train)
        train.register_output(y_train)
        trained_model = .....
        train.save_model(trained_model)

    """
    project_full_name = current_project_full_name()
    if not project_full_name:
        raise Exception(
            "Missing current project name, please do 'layer.init(\"account-name/project-name\")'"
            + " or 'layer.init(\"project-name\")'"
        )
    config = asyncio_run_in_thread(ConfigManager().refresh())
    with LayerClient(config.client, logger).init() as client:
        train = Train(
            layer_client=client,
            name=name,
            version=version,
            train_id=train_id,
            project_full_name=project_full_name,
        )
        with train:
            yield train


def init(
    project_name: str,
    fabric: Optional[str] = None,
    pip_packages: Optional[List[str]] = None,
    pip_requirements_file: Optional[str] = None,
) -> Project:
    """
    :param project_name: Name of the project to be initialized.
    :param fabric: Default fabric to use for current project when running code in Layer's backend.
    :param pip_packages: List of packages to install in Layer backend when running code for this project.
    :param pip_requirements_file: File name with list of packages to install in Layer backend when running code for this project.
    :return: Project object.

    Initializes a project with given name both locally and remotely.

    If the project does not exist in the Layer backend, then it will be automatically created.
    Additional settings can be passed in as parameters: namely fabric and packages to install.
    These settings are only applied to runs initiated from the current local environment.

    Other users can set different settings for the same project in their own local environments
    via ``layer.init``, and those will not impact your runs.

    Note that only one of ``pip_packages`` or ``pip_requirements_file`` should be provided.

    .. code-block:: python

        # Initialize new project and create it in Layer backend if it does not exist
        project = layer.init("my_project_name", fabric="x-small")
    """
    _check_latest_version()

    if pip_packages and pip_requirements_file:
        raise ValueError(
            "either pip_requirements_file or pip_packages should be provided, not both"
        )
    layer_config = asyncio_run_in_thread(ConfigManager().refresh())

    project_full_name = _get_project_full_name(layer_config, project_name)

    reset_to(project_full_name)

    init_project_runner = InitProjectRunner(project_full_name, logger=logger)
    fabric_to_set = Fabric(fabric) if fabric else None
    return init_project_runner.setup_project(
        fabric=fabric_to_set,
        pip_packages=pip_packages,
        pip_requirements_file=pip_requirements_file,
    )


def run(functions: List[Any], debug: bool = False) -> Run:
    """
    :param functions: List of decorated functions to run in the Layer backend.
    :param debug: Stream logs to console from infra executing the project remotely.
    :return: An object with information on the run triggered, unless there is an initialization error.

    Runs the specified functions on Layer infrastructure.

    Running the project does the following:
    - Adds new versions of decorated functions to the Layer web UI.
    - Runs any logs and stores them.

    .. code-block:: python

        import pandas as pd
        from layer
        from layer.decorators import dataset

        # Define a new function for dataset generation
        @dataset("my_dataset_name")
        def create_my_dataset():
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            dataframe = pd.DataFrame(data, columns=["Id", "Product", "Price"])
            return dataframe

        # Initialize current project locally and remotely (if it doesn't exist)
        project = layer.init("my_project_name", fabric="x-small")

        # Run the dataset generation code using Layer's backend compute resources
        run = layer.run([create_my_dataset])
        # run = layer.run([create_my_dataset], debug=True)  # Stream logs to console
    """
    _check_python_version()
    _ensure_all_functions_are_decorated(functions)

    layer_config = asyncio_run_in_thread(ConfigManager().refresh())

    project_full_name = _get_project_full_name(layer_config, get_current_project_name())
    project_runner = ProjectRunner(config=layer_config)
    run = project_runner.with_functions(project_full_name, functions)
    run = project_runner.run(run, debug=debug)

    _make_notebook_links_open_in_new_tab()

    return run


def _get_project_full_name(
    layer_config: Config, user_input_project_name: str
) -> ProjectFullName:
    """
    Will first try to extract account_name/project_name from :user_input_project_name

    If no account name can be extracted, will try to get it from global context.

    If that too fails, will fetch the personal account.
    """
    parts = user_input_project_name.split("/")
    account_name: Optional[str]
    project_name: str
    if len(parts) == 1:
        project_name = user_input_project_name
        with LayerClient(layer_config.client, logger).init() as client:
            account_name = client.account.get_my_account().name
    else:
        account_name = parts[0]
        project_name = parts[1]

    return ProjectFullName(
        project_name=project_name,
        account_name=account_name,
    )


# Normally, Colab/IPython opens links as an IFrame. One can open them as new tabs through the right-click menu or using shift+click.
# However, we would like to change the default behavior to always open on a new page.
def _make_notebook_links_open_in_new_tab() -> None:
    def is_notebook() -> bool:
        try:
            shell = get_ipython().__class__.__name__  # type: ignore
            if shell == "ZMQInteractiveShell":
                return True  # Jupyter notebook or qtconsole
            elif shell == "TerminalInteractiveShell":
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    if is_notebook():
        from IPython.display import Javascript, display  # type: ignore

        # Colab
        display(
            Javascript(
                """
document.querySelectorAll("#output-area a").forEach(function(a){
    a.setAttribute('target', '_blank');
})
        """
            )
        )

        # Jupyter
        display(
            Javascript(
                """
document.querySelectorAll(".output a").forEach(function(a){
    a.setAttribute('target', '_blank');
})
        """
            )
        )


def _check_latest_version() -> None:
    if has_shown_update_message():
        return

    import luddite  # type: ignore
    import pkg_resources

    latest_version = luddite.get_version_pypi("layer")
    current_version = pkg_resources.get_distribution("layer").version
    if current_version != latest_version:
        print(
            f"You are using the version {current_version} but the latest version is {latest_version}, please upgrade with 'pip install --upgrade layer'"
        )
    set_has_shown_update_message(True)


def _check_python_version() -> None:
    if has_shown_python_version_message():
        return

    import platform

    major, minor, _ = platform.python_version_tuple()

    if major != "3" or minor not in ["7", "8"]:
        print(
            f"You are using the Python version {platform.python_version()} but layer requires Python 3.7.x or 3.8.x"
        )
    set_has_shown_python_version_message(True)


def _ensure_all_functions_are_decorated(functions: List[Callable[..., Any]]) -> None:
    for f in functions:
        if not hasattr(f, "layer"):
            raise ConfigError(
                'Either @dataset(name="...") or @model(name="...") is required for each function. '
                "Add @dataset or @model decorator to your functions to run them in Layer."
            )
        settings: LayerSettings = f.layer  # type: ignore
        settings.validate()


def log(
    data: Dict[
        str,
        Union[
            str,
            float,
            bool,
            int,
            Dict[str, Any],
            "pandas.DataFrame",
            "PIL.Image.Image",
            "matplotlib.figure.Figure",
            "matplotlib.axes._subplots.AxesSubplot",
            ModuleType,
            Path,
        ],
    ],
    step: Optional[int] = None,
) -> None:
    """
    :param data: A dictionary in which each key is a string tag (i.e. name/id). The value can have different types. See examples below for more details.
    :param step: An optional non-negative integer that associates data with a particular step (epoch). This only takes effect if the logged data is to be associated with a model train (and *not* with a dataset build), and the data is either a number or an image.
    :return: None

    Logs arbitrary data associated with a model train or a dataset build into Layer backend.

    This function can only be run inside functions decorated with ``@model`` or ``@dataset``. The logged data can then be discovered and analyzed through the Layer UI.

    We support Python primitives, images, tables and plots to enable better experiment tracking and version comparison.

    **Python Primitives**

    You can log Python primitive types for your model train parameters, metrics or KPIs

    Accepted Types:
    ``str``,
    ``float``,
    ``bool``,
    ``int``

    **Images**

    You can log images to track inputs, outputs, detections, activations and more. We support JPEG and PNG formats.

    Accepted Types:
    ``PIL.Image.Image``,
    ``path.Path``

    **Charts**

    You can track your metrics in detail with charts

    Accepted Types:
    ``matplotlib.figure.Figure``,
    ``matplotlib.pyplot``,
    ``matplotlib.axes._subplots.AxesSubplot``,
    ``ModuleType`` (only for the matplotlib module, for convenience)

    **Tables**

    You can log dataframes to display and analyze your tabular data.

    Accepted Types:
    ``pandas.DataFrame``
    ``dict`` (the key should be a string. The value either needs to be a primitive type or it will be converted to str)

    .. code-block:: python

        import layer
        import matplotlib.pyplot as plt
        import pandas as pd
        from layer.decorators import dataset, model

        # Define a new function for dataset generation
        @dataset("my_dataset_name")
        def create_my_dataset():
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            dataframe = pd.DataFrame(data, columns=["Id", "Product", "Price"])

            t = np.arange(0.0, 2.0, 0.01)
            s = 1 + np.sin(2 * np.pi * t)
            fig, ax = plt.subplots()

            layer.log({
                "my-str-tag": "any text",
                "my-int-tag": 123, # any number
                "foo-bool": True,
                "some-sample-dataframe-tag": ..., # Pandas data frame
                "some-local-image-file": Path.home() / "images/foo.png",
                "some-matplot-lib-figure": fig, # You could alternatively just passed plt as well, and Layer would just get the current/active figure.
            })

            layer.log({
                "another-tag": "you can call layer.log multiple times"
            })

            return dataframe

        @model("my_model")
        def create_my_model():
            # everything that can be logged for a dataset (as shown above), can also be logged for a model too.

            # In addition to those, if you have a multi-step (i.e. multi-epoch) algorithm, you can associate a metric with the step.
            # These will be rendered on a graph inside the Layer UI.
            for i in range(1000):
                some_result, accuracy = some_algo(some_result)
                layer.log({
                    "Model accuracy": accuracy,
                }, step=i)

        create_my_dataset()
        create_my_model()
    """
    active_context = get_active_context()
    if not active_context:
        raise RuntimeError(
            "Data logging only allowed inside functions decorated with @model or @dataset"
        )
    train = active_context.train()
    train_id = train.get_id() if train is not None else None
    dataset_build = active_context.dataset_build()
    dataset_build_id = dataset_build.id if dataset_build is not None else None
    layer_config = asyncio_run_in_thread(ConfigManager().refresh())
    with LayerClient(layer_config.client, logger).init() as client:
        log_data_runner = LogDataRunner(
            client=client,
            train_id=train_id,
            dataset_build_id=dataset_build_id,
            logger=logger,
        )
        log_data_runner.log(data=data, epoch=step)


def clear_cache() -> None:
    """
    Clear all cached objects fetched by the Layer SDK on this machine.
    The Layer SDK locally stores all datasets and models by default on your computer.
    When you fetch a dataset with :func:``layer.get_dataset``, or load the model with ``layer.get_model``,
    the first call will fetch the artifact from remote storage,
    but subsequent calls will read it from the local disk.
    """
    Cache().clear()



import datetime
from typing import List, Optional

import humanize  # type: ignore
import polling  # type: ignore
from layerapi.api.entity.run_pb2 import Run
from rich.console import Console, RenderableType
from rich.progress import Progress, ProgressColumn, Task
from rich.spinner import Spinner
from rich.style import StyleType
from rich.table import Column
from rich.text import Text, TextType

from layer.contracts.runs import GetRunsFunction

from .asset_column import AssetColumn


# Taken from https://github.com/willmcgugan/rich/blob/6f09ae226c26a2be52e3214ee93e6d704756d282/rich/progress.py#L211
class SpinnerColumn(ProgressColumn):
    """A column with a 'spinner' animation.
    Args:
        spinner_name (str, optional): Name of spinner animation. Defaults to "dots".
        style (StyleType, optional): Style of spinner. Defaults to "progress.spinner".
        speed (float, optional): Speed factor of spinner. Defaults to 1.0.
        finished_text (TextType, optional): Text used when task is finished. Defaults to " ".
    """

    def __init__(
        self,
        spinner_name: str = "dots",
        style: Optional[StyleType] = "progress.spinner",
        speed: float = 1.0,
        finished_text: TextType = " ",
        table_column: Optional[Column] = None,
    ):
        self.spinner = Spinner(spinner_name, style=style, speed=speed)
        self.finished_text = (
            Text.from_markup(finished_text)
            if isinstance(finished_text, str)
            else finished_text
        )
        super().__init__(table_column=table_column)

    def set_spinner(
        self,
        spinner_name: str,
        spinner_style: Optional[StyleType] = "progress.spinner",
        speed: float = 1.0,
    ) -> None:
        """Set a new spinner.
        Args:
            spinner_name (str): Spinner name, see python -m rich.spinner.
            spinner_style (Optional[StyleType], optional): Spinner style. Defaults to "progress.spinner".
            speed (float, optional): Speed factor of spinner. Defaults to 1.0.
        """
        self.spinner = Spinner(spinner_name, style=spinner_style, speed=speed)

    def render(self, task: Task) -> RenderableType:
        text = (
            self.finished_text
            if task.finished
            else self.spinner.render(task.get_time())
        )
        return text


def get_progress_ui() -> Progress:
    return Progress(
        SpinnerColumn(finished_text=":white_heavy_check_mark:"),
        AssetColumn(),
        console=Console(),
    )


def watch_get_runs(console: Console, get_runs_fn: GetRunsFunction) -> None:
    def clean_get_print() -> None:
        console.clear()
        print_runs(get_runs_fn())

    polling.poll(
        clean_get_print,
        step=5,
        poll_forever=True,
    )


def print_runs(runs: List[Run]) -> None:
    print(
        f"{'ID':<40} {'PROJECT':<60} {'STATUS':<24} {'CREATED TIME':<24} {'DURATION'}"
    )
    for run in runs:
        print(_to_run_string(run))


def _to_run_string(run: Run) -> str:
    created_time = run.created_time.ToDatetime()
    created_time_formatted = created_time.strftime("%Y-%m-%d %H:%M:%S")
    humanized_duration_str = humanize.naturaldelta(
        datetime.timedelta(seconds=run.duration.seconds)
    )
    status = str(Run.Status.Name(run.run_status))
    return f"{run.id.value:<40} {run.project_name:<60} {status[len('STATUS_'):]:<24} {created_time_formatted:<24} {humanized_duration_str}"



import os
import posixpath
from mimetypes import guess_type
from pathlib import Path
from typing import Optional

from layerapi.api.value.aws_credentials_pb2 import AwsCredentials
from layerapi.api.value.s3_path_pb2 import S3Path
from yarl import URL

from layer.contracts.tracker import ResourceTransferState


class S3Util:
    @staticmethod
    def download_dir(
        local_dir: Path,
        s3_path: S3Path,
        credentials: Optional[AwsCredentials] = None,
        *,
        endpoint_url: Optional[URL] = None,
        state: ResourceTransferState,
    ) -> None:
        import boto3

        s3_kwargs = {"endpoint_url": endpoint_url and str(endpoint_url)}

        if credentials:
            s3_kwargs.update(
                {
                    "aws_access_key_id": credentials.access_key_id,
                    "aws_secret_access_key": credentials.secret_access_key,
                    "aws_session_token": credentials.session_token,
                }
            )
        s3 = boto3.resource("s3", **s3_kwargs)  # type: ignore
        bucket = s3.Bucket(s3_path.bucket)

        to_download = []
        total_num_files = 0
        total_file_bytes = 0
        for obj in bucket.objects.filter(Prefix=s3_path.key):
            target = os.path.join(local_dir, os.path.relpath(obj.key, s3_path.key))
            if not os.path.exists(os.path.dirname(target)):
                os.makedirs(os.path.dirname(target))
            if obj.key[-1] == "/":
                continue

            total_num_files += 1
            total_file_bytes += obj.size
            to_download.append(
                {
                    "args": [obj.key, target],
                    "kwargs": {
                        "Callback": state.increment_transferred_resource_size_bytes
                    },
                }
            )

        state.total_num_files = total_num_files
        state.total_resource_size_bytes = total_file_bytes
        for function_params in to_download:
            bucket.download_file(*function_params["args"], **function_params["kwargs"])

    @staticmethod
    def upload_dir(
        local_dir: Path,
        credentials: AwsCredentials,
        s3_path: S3Path,
        *,
        endpoint_url: Optional[URL] = None,
        state: Optional[ResourceTransferState] = None,
    ) -> None:
        import boto3
        from mlflow.utils.file_utils import relative_path_to_artifact_path

        if not state:
            state = ResourceTransferState()

        s3_kwargs = {"endpoint_url": endpoint_url and str(endpoint_url)}

        if credentials:
            s3_kwargs.update(
                {
                    "aws_access_key_id": credentials.access_key_id,
                    "aws_secret_access_key": credentials.secret_access_key,
                    "aws_session_token": credentials.session_token,
                }
            )

        s3_client = boto3.client("s3", **s3_kwargs)  # type: ignore
        dest_path = s3_path.key
        absolute_path = os.path.abspath(local_dir)
        total_file_bytes = 0
        total_num_files = 0
        to_upload = []
        for (root, _, filenames) in os.walk(absolute_path):
            upload_path = dest_path
            if root != absolute_path:
                rel_path = os.path.relpath(root, absolute_path)
                rel_path = relative_path_to_artifact_path(rel_path)
                upload_path = posixpath.join(dest_path, rel_path)
            for f in filenames:
                filename = os.path.join(root, f)
                total_num_files += 1
                total_file_bytes += os.path.getsize(filename)
                extra_args = {}
                guessed_type, guessed_encoding = guess_type(filename)
                if guessed_type is not None:
                    extra_args["ContentType"] = guessed_type
                if guessed_encoding is not None:
                    extra_args["ContentEncoding"] = guessed_encoding
                to_upload.append(
                    {
                        "Filename": filename,
                        "Bucket": s3_path.bucket,
                        "Key": posixpath.join(upload_path, f),
                        "ExtraArgs": extra_args,
                    }
                )
        state.total_num_files = total_num_files
        state.total_resource_size_bytes = total_file_bytes
        for upload_file_params in to_upload:
            s3_client.upload_file(
                **upload_file_params,
                Callback=state.increment_transferred_resource_size_bytes,
            )
            state.increment_num_transferred_files(1)

    @staticmethod
    def create_dir(
        credentials: AwsCredentials,
        bucket: str,
        key: str,
        *,
        endpoint_url: Optional[URL] = None,
    ) -> None:
        import boto3

        s3_kwargs = {"endpoint_url": endpoint_url and str(endpoint_url)}

        if credentials:
            s3_kwargs.update(
                {
                    "aws_access_key_id": credentials.access_key_id,
                    "aws_secret_access_key": credentials.secret_access_key,
                    "aws_session_token": credentials.session_token,
                }
            )

        s3_client = boto3.client("s3", **s3_kwargs)  # type: ignore
        s3_client.put_object(Bucket=bucket, Body="", Key=key)



from logging import Logger
from pathlib import Path
from uuid import UUID

from layerapi.api.entity.model_train_status_pb2 import ModelTrainStatus
from layerapi.api.ids_pb2 import ModelTrainId

from layer.clients.model_catalog import ModelCatalogClient
from layer.exceptions.exceptions import LayerFailedAssertionsException
from layer.exceptions.status_report import (
    ExecutionStatusReport,
    ExecutionStatusReportFactory,
    PythonExecutionStatusReport,
)
from layer.training.runtime.common import update_train_status


class ModelTrainFailureReporter:
    model_catalog_client: ModelCatalogClient
    logger: Logger
    train_id: UUID
    source_folder: Path

    def __init__(
        self,
        model_catalog_client: ModelCatalogClient,
        logger: Logger,
        train_id: UUID,
        source_folder: Path,
    ) -> None:
        self.model_catalog_client = model_catalog_client
        self.logger = logger
        self.train_id = train_id
        self.source_folder = source_folder

    def report_failure(self, stage: str, failure_exc: Exception) -> None:
        # Check to only keep cause for inner most exception, as __exit__
        # catches SystemExit exceptions, thus without this if a chain of status updates
        # could be triggered with the outer most exception message overriding inner ones
        existing_status = self._get_train_status()
        if existing_status != ModelTrainStatus.TRAIN_STATUS_FAILED:
            report: ExecutionStatusReport
            if isinstance(failure_exc, LayerFailedAssertionsException):
                report = failure_exc.to_status_report()
            else:
                report = PythonExecutionStatusReport.from_exception(
                    failure_exc, self.source_folder
                )
            update_train_status(
                self.model_catalog_client,
                self.train_id,
                ModelTrainStatus.TRAIN_STATUS_FAILED,
                self.logger,
                info=ExecutionStatusReportFactory.to_json(report),
            )

    def _get_train_status(self) -> "ModelTrainStatus.TrainStatus.V":
        return self.model_catalog_client.get_model_train(
            ModelTrainId(value=str(self.train_id))
        ).train_status.train_status
